//this file is part of eMule
//Copyright (C)2002-2007 Merkur ( strEmail.Format("%s@%s", "devteam", "emule-project.net") / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#include "stdafx.h"
#include "emule.h"
#include "ClientList.h"
#include "otherfunctions.h"
#include "Kademlia/Kademlia/kademlia.h"
#include "Kademlia/Kademlia/prefs.h"
#include "Kademlia/Kademlia/search.h"
#include "Kademlia/Kademlia/searchmanager.h"
#include "Kademlia/routing/contact.h"
#include "Kademlia/net/kademliaudplistener.h"
#include "kademlia/utils/UInt128.h"
#include "LastCommonRouteFinder.h"
#include "UploadQueue.h"
#include "DownloadQueue.h"
#include "UpDownClient.h"
#include "ClientCredits.h"
#include "ListenSocket.h"
#include "Opcodes.h"
#include "Sockets.h"
#include "emuledlg.h"
#include "TransferWnd.h"
#include "serverwnd.h"
#include "Log.h"
#include "packets.h"
#include "Statistics.h"
#include "Preferences.h" // X-Ray :: DropSystem
#include "server.h" // X-Ray :: ReAskSrcAfterIPChange
#include "Addons/Argos/Argos.h" // X-Ray :: Argos

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


CClientList::CClientList(){
	// X-Ray :: RelaxOnStartUp :: Start
	/*
	m_dwLastBannCleanUp = 0;
	m_dwLastTrackedCleanUp = 0;
	m_dwLastClientCleanUp = 0;
	*/
	const DWORD cur_tick = ::GetTickCount();
	m_dwLastBannCleanUp = cur_tick + CLIENTBANTIME; 	
	m_dwLastTrackedCleanUp = cur_tick + TRACKED_CLEANUP_TIME;
	m_dwLastClientCleanUp = cur_tick + CLIENTLIST_CLEANUP_TIME;
	// X-Ray :: RelaxOnStartUp :: End

	// X-Ray :: DropSystem :: Start
	m_dwLastCleanUpDontAskThisIP = 0;
	m_DontAskThisIPList.InitHashTable(22229);
	// X-Ray :: DropSystem :: End

	m_nBuddyStatus = Disconnected;
	// X-Ray :: DropSystem :: Start
	/*
	m_bannedList.InitHashTable(331);
	*/
	m_bannedList.InitHashTable(2011);
	// X-Ray :: DropSystem :: End
	m_trackedClientsList.InitHashTable(2011);
	m_globDeadSourceList.Init(true);
	m_pBuddy = NULL;
}

CClientList::~CClientList(){
	RemoveAllTrackedClients();
}

void CClientList::GetStatistics(uint32 &ruTotalClients, int stats[NUM_CLIENTLIST_STATS], 
								CMap<uint32, uint32, uint32, uint32>& clientVersionEDonkey, 
								CMap<uint32, uint32, uint32, uint32>& clientVersionEDonkeyHybrid, 
								CMap<uint32, uint32, uint32, uint32>& clientVersionEMule, 
								CMap<uint32, uint32, uint32, uint32>& clientVersionAMule)
{
	ruTotalClients = list.GetCount();
	memset(stats, 0, sizeof(stats[0]) * NUM_CLIENTLIST_STATS);

	for (POSITION pos = list.GetHeadPosition(); pos != NULL; )
	{
		const CUpDownClient* cur_client = list.GetNext(pos);

		if (cur_client->HasLowID())
			stats[14]++;

		switch (cur_client->GetClientSoft())
		{
		case SO_EMULE:
		case SO_OLDEMULE:
			stats[2]++;
			clientVersionEMule[cur_client->GetVersion()]++;
			break;

		case SO_EDONKEYHYBRID : 
			stats[4]++;
			clientVersionEDonkeyHybrid[cur_client->GetVersion()]++;
			break;

		case SO_AMULE:
			stats[10]++;
			clientVersionAMule[cur_client->GetVersion()]++;
			break;

		case SO_EDONKEY:
			stats[1]++;
			clientVersionEDonkey[cur_client->GetVersion()]++;
			break;

		case SO_MLDONKEY:
			stats[3]++;
			break;

		case SO_SHAREAZA:
			stats[11]++;
			break;

		// all remaining 'eMule Compatible' clients
		// X-Ray :: EnhancedClientRecognization :: Start
		case SO_HYDRANODE:
		case SO_EMULEPLUS:
		case SO_TRUSTYFILES:
		// X-Ray :: EnhancedClientRecognization :: End
		case SO_CDONKEY:
		case SO_XMULE:
		case SO_LPHANT:
			stats[5]++;
			break;

		default:
			stats[0]++;
			break;
		}

		if (cur_client->Credits() != NULL)
		{
			switch (cur_client->Credits()->GetCurrentIdentState(cur_client->GetIP()))
			{
			case IS_IDENTIFIED:
				stats[12]++;
				break;
			case IS_IDFAILED:
			case IS_IDNEEDED:
			case IS_IDBADGUY:
				stats[13]++;
				break;
			}
		}

		if (cur_client->GetDownloadState()==DS_ERROR || cur_client->GetUploadState()==US_ERROR)
			stats[6]++; // Error

		switch (cur_client->GetUserPort())
		{
		case 4662:
			stats[8]++; // Default Port
			break;
		default:
			stats[9]++; // Other Port
		}

		// Network client stats
		if (cur_client->GetServerIP() && cur_client->GetServerPort())
		{
			stats[15]++;		// eD2K
			if(cur_client->GetKadPort())
			{
				stats[17]++;	// eD2K/Kad
				stats[16]++;	// Kad
			}
		}
		else if (cur_client->GetKadPort())
			stats[16]++;		// Kad
		else
			stats[18]++;		// Unknown
	}
}

// X-Ray :: ModID :: Start
void CClientList::GetModStatistics(CRBMap<uint32, CRBMap<CString, uint32>* > *clientMods){
	if (!clientMods)
		return;
	clientMods->RemoveAll();
	POSITION pos1, pos2;

	for (pos1 = list.GetHeadPosition();( pos2 = pos1 ) != NULL;){
		list.GetNext(pos1);
		CUpDownClient* cur_client =	list.GetAt(pos2);

		switch (cur_client->GetClientSoft()) {
		case SO_EMULE   :
		case SO_OLDEMULE:
			break;
		default:
			continue;
		}

		CRBMap<CString, uint32> *versionMods;

		if (!clientMods->Lookup(cur_client->GetVersion(), versionMods)){
			versionMods = new CRBMap<CString, uint32>;
			versionMods->RemoveAll();
			clientMods->SetAt(cur_client->GetVersion(), versionMods);
		}

		uint32 count;

		if (!versionMods->Lookup(cur_client->GetClientModVer(), count))
			count = 1;
		else
			count++;

		versionMods->SetAt(cur_client->GetClientModVer(), count);
	}
}

void CClientList::ReleaseModStatistics(CRBMap<uint32, CRBMap<CString, uint32>* > *clientMods){
	if (!clientMods)
		return;
	POSITION pos = clientMods->GetHeadPosition();
	while(pos != NULL)
	{
		uint32 version;
		CRBMap<CString, uint32> *versionMods;
		clientMods->GetNextAssoc(pos, version, versionMods);
		delete versionMods;
	}
	clientMods->RemoveAll();
}
// X-Ray :: ModID :: End

void CClientList::AddClient(CUpDownClient* toadd, bool bSkipDupTest)
{
	// skipping the check for duplicate list entries is only to be done for optimization purposes, if the calling
	// function has ensured that this client instance is not already within the list -> there are never duplicate
	// client instances in this list.
	if (!bSkipDupTest){
		if(list.Find(toadd))
			return;
	}
	theApp.emuledlg->transferwnd->clientlistctrl.AddClient(toadd);
	list.AddTail(toadd);
}

// ZZ:UploadSpeedSense -->
bool CClientList::GiveClientsForTraceRoute() {
    // this is a host that lastCommonRouteFinder can use to traceroute
    return theApp.lastCommonRouteFinder->AddHostsToCheck(list);
}
// ZZ:UploadSpeedSense <--

void CClientList::RemoveClient(CUpDownClient* toremove, LPCTSTR pszReason){
	POSITION pos = list.Find(toremove);
	if (pos){
        theApp.uploadqueue->RemoveFromUploadQueue(toremove, CString(_T("CClientList::RemoveClient: ")) + pszReason);
		theApp.uploadqueue->RemoveFromWaitingQueue(toremove);

		// X-Ray :: SUQWT :: Start
		if ( toremove != NULL && toremove->Credits() != NULL) {
			toremove->Credits()->ClearWaitStartTime();
		}
		// X-Ray :: SUQWT :: End

		theApp.downloadqueue->RemoveSource(toremove);
		theApp.emuledlg->transferwnd->clientlistctrl.RemoveClient(toremove);
		RemoveFromKadList(toremove);
		list.RemoveAt(pos);
	}
}

void CClientList::DeleteAll(){
	theApp.uploadqueue->DeleteAll();
	theApp.downloadqueue->DeleteAll();

	// X-Ray :: LoopImprovement :: Start
	/*
	POSITION pos1, pos2;
	for (pos1 = list.GetHeadPosition();( pos2 = pos1 ) != NULL;){
		list.GetNext(pos1);
		CUpDownClient* cur_client =	list.GetAt(pos2);
		list.RemoveAt(pos2);
		delete cur_client; // recursiv: this will call RemoveClient
	}
	*/
	while (!list.IsEmpty())
		delete list.RemoveHead(); // WiZaRd Fix
	// X-Ray :: LoopImprovement :: End
}

// X-Ray :: BetterClientAttaching :: Start
/*
bool CClientList::AttachToAlreadyKnown(CUpDownClient** client, CClientReqSocket* sender){
*/
bool CClientList::AttachToAlreadyKnown(CUpDownClient** client, CClientReqSocket* sender, EAttachMode AttachMode){ 
// X-Ray :: BetterClientAttaching :: End
	// X-Ray :: LoopImprovement :: Start
	/*
	POSITION pos1, pos2;
	*/
	// X-Ray :: LoopImprovement :: End
	CUpDownClient* tocheck = (*client);
	CUpDownClient* found_client = NULL;
	CUpDownClient* found_client2 = NULL;

	// X-Ray :: BetterClientAttaching :: Start
	/*
	for (pos1 = list.GetHeadPosition(); (pos2 = pos1) != NULL; ){
		list.GetNext(pos1);
		CUpDownClient* cur_client =	list.GetAt(pos2);
		if (tocheck->Compare(cur_client,false)){ //matching userhash
			found_client2 = cur_client;
		}
		if (tocheck->Compare(cur_client,true)){	 //matching IP
			found_client = cur_client;
			break;
		}
	}
	*/
	if(AttachMode == ATTACH_ANSWER){
		// X-Ray :: LoopImprovement :: Start
		/*
		for (pos1 = list.GetHeadPosition(); (pos2 = pos1) != NULL; ){
			list.GetNext(pos1);
			CUpDownClient* cur_client =	list.GetAt(pos2);
		*/
		for (POSITION pos = list.GetHeadPosition(); pos; ){
			CUpDownClient* cur_client = list.GetNext(pos);
		// X-Ray :: LoopImprovement :: End
			if (tocheck->Compare(cur_client,false) && !tocheck->Compare(cur_client,true)){ //matching userhash & different IP
				found_client = cur_client;
				break;
			}
		}
	}
	else {
		// X-Ray :: LoopImprovement :: Start
		/*
		for (pos1 = list.GetHeadPosition(); (pos2 = pos1) != NULL; ){
			list.GetNext(pos1);
			CUpDownClient* cur_client =	list.GetAt(pos2);
		*/
		for (POSITION pos = list.GetHeadPosition(); pos; ){
			CUpDownClient* cur_client = list.GetNext(pos);
		// X-Ray :: LoopImprovement :: End
			if (tocheck->Compare(cur_client,false)) //matching userhash
				found_client2 = cur_client;
			if (tocheck->Compare(cur_client,true)){	 //matching IP
				found_client = cur_client;
				break;
			}
		}
		if (found_client == NULL)
			found_client = found_client2;
	}
	// X-Ray :: BetterClientAttaching :: End
	if (found_client == NULL)
		found_client = found_client2;

	if (found_client != NULL){
		if (tocheck == found_client){
			//we found the same client instance (client may have sent more than one OP_HELLO). do not delete that client!
			// X-Ray :: BetterClientAttaching :: Start
			//here the cleint already have processed the hello no need to process it 2 times
			/*
			return true;
			*/
			return AttachMode != ATTACH_HALLO;
			// X-Ray :: BetterClientAttaching :: End
		}
		if (sender){
			if (found_client->socket){
				if (found_client->socket->IsConnected() 
					&& (found_client->GetIP() != tocheck->GetIP() || found_client->GetUserPort() != tocheck->GetUserPort() ) )
				{
					// if found_client is connected and has the IS_IDENTIFIED, it's safe to say that the other one is a bad guy
					if (found_client->Credits() && found_client->Credits()->GetCurrentIdentState(found_client->GetIP()) == IS_IDENTIFIED){
						if (thePrefs.GetLogBannedClients())
							AddDebugLogLine(false, _T("Clients: %s (%s), Banreason: Userhash invalid"), tocheck->GetUserName(), ipstr(tocheck->GetConnectIP()));
						// X-Ray :: Argos :: Start
						if(thePrefs.UseArgosSystem())
							theApp.argos->DoArgos(tocheck->GetConnectIP(),AR_HASHTHIEF,TRUE);
						else
						// X-Ray :: Argos :: End
							tocheck->Ban(_T("Userhash invalid")); // X-Ray :: MOD
						return false;
					}
	
					//IDS_CLIENTCOL Warning: Found matching client, to a currently connected client: %s (%s) and %s (%s)
					if (thePrefs.GetLogBannedClients())
						AddDebugLogLine(true,GetResString(IDS_CLIENTCOL), tocheck->GetUserName(), ipstr(tocheck->GetConnectIP()), found_client->GetUserName(), ipstr(found_client->GetConnectIP()));
					return false;
				}
				found_client->socket->client = 0;
				found_client->socket->Safe_Delete();
			}
			found_client->socket = sender;
			tocheck->socket = 0;
		}
		*client = 0;
		delete tocheck;
		*client = found_client;
		return true;
	}
	return false;
}

CUpDownClient* CClientList::FindClientByIP(uint32 clientip, UINT port) const
{
	for (POSITION pos = list.GetHeadPosition(); pos != NULL;)
	{
		CUpDownClient* cur_client = list.GetNext(pos);
		if (cur_client->GetIP() == clientip && cur_client->GetUserPort() == port)
			return cur_client;
	}
	return 0;
}

CUpDownClient* CClientList::FindClientByUserHash(const uchar* clienthash) const
{
	for (POSITION pos = list.GetHeadPosition(); pos != NULL;)
	{
		CUpDownClient* cur_client = list.GetNext(pos);
		if (!md4cmp(cur_client->GetUserHash() ,clienthash))
			return cur_client;
	}
	return 0;
}

CUpDownClient* CClientList::FindClientByIP(uint32 clientip) const
{
	for (POSITION pos = list.GetHeadPosition(); pos != NULL;)
	{
		CUpDownClient* cur_client = list.GetNext(pos);
		if (cur_client->GetIP() == clientip)
			return cur_client;
	}
	return 0;
}

CUpDownClient* CClientList::FindClientByIP_UDP(uint32 clientip, UINT nUDPport) const
{
	for (POSITION pos = list.GetHeadPosition(); pos != NULL;)
	{
		CUpDownClient* cur_client = list.GetNext(pos);
		if (cur_client->GetIP() == clientip && cur_client->GetUDPPort() == nUDPport)
			return cur_client;
	}
	return 0;
}

CUpDownClient* CClientList::FindClientByUserID_KadPort(uint32 clientID, uint16 kadPort) const
{
	for (POSITION pos = list.GetHeadPosition(); pos != NULL;)
	{
		CUpDownClient* cur_client = list.GetNext(pos);
		if (cur_client->GetUserIDHybrid() == clientID && cur_client->GetKadPort() == kadPort)
			return cur_client;
	}
	return 0;
}

CUpDownClient* CClientList::FindClientByIP_KadPort(uint32 ip, uint16 port) const
{
	for (POSITION pos = list.GetHeadPosition(); pos != NULL;)
	{
		CUpDownClient* cur_client = list.GetNext(pos);
		if (cur_client->GetIP() == ip && cur_client->GetKadPort() == port)
			return cur_client;
	}
	return 0;
}

CUpDownClient* CClientList::FindClientByServerID(uint32 uServerIP, uint32 uED2KUserID) const
{
	uint32 uHybridUserID = ntohl(uED2KUserID);
	for (POSITION pos = list.GetHeadPosition(); pos != NULL;)
	{
		CUpDownClient* cur_client =	list.GetNext(pos);
		if (cur_client->GetServerIP() == uServerIP && cur_client->GetUserIDHybrid() == uHybridUserID)
			return cur_client;
	}
	return 0;
}


///////////////////////////////////////////////////////////////////////////////
// Banned clients

void CClientList::AddBannedClient(uint32 dwIP){
	m_bannedList.SetAt(dwIP, ::GetTickCount());
}

bool CClientList::IsBannedClient(uint32 dwIP) const
{
	uint32 dwBantime;
	if (m_bannedList.Lookup(dwIP, dwBantime)){
		// X-Ray :: Argos :: Start
		/*
		if (dwBantime + CLIENTBANTIME > ::GetTickCount())
		*/
		if (dwBantime + thePrefs.GetArgosGenericBanTimeMs() > ::GetTickCount())
		// X-Ray :: Argos :: End
			return true;
	}
	return false; 
}

void CClientList::RemoveBannedClient(uint32 dwIP){
	m_bannedList.RemoveKey(dwIP);
}

void CClientList::RemoveAllBannedClients(){
	m_bannedList.RemoveAll();
}


///////////////////////////////////////////////////////////////////////////////
// Tracked clients

void CClientList::AddTrackClient(CUpDownClient* toadd){
	CDeletedClient* pResult = 0;
	if (m_trackedClientsList.Lookup(toadd->GetIP(), pResult)){
		pResult->m_dwInserted = ::GetTickCount();
		for (int i = 0; i != pResult->m_ItemsList.GetCount(); i++){
			if (pResult->m_ItemsList[i].nPort == toadd->GetUserPort()){
				// X-Ray :: ExtenedCreditTableArragement :: Start
				/*
				// already tracked, update
				pResult->m_ItemsList[i].pHash = toadd->Credits();
				*/
				// don't keep a track of the credit-pointer, but of the hash
				md4cpy(pResult->m_ItemsList[i].pHash, toadd->GetUserHash());
				// X-Ray :: ExtenedCreditTableArragement :: End
				return;
			}
		}
		// X-Ray :: ExtenedCreditTableArragement :: Start
		/*
		PORTANDHASH porthash = { toadd->GetUserPort(), toadd->Credits()};
		*/
		PORTANDHASH porthash;
		porthash.nPort = toadd->GetUserPort();
		md4cpy(porthash.pHash, toadd->GetUserHash());
		// X-Ray :: ExtenedCreditTableArragement :: End
		pResult->m_ItemsList.Add(porthash);
	}
	else{
		m_trackedClientsList.SetAt(toadd->GetIP(), new CDeletedClient(toadd));
	}
}

// true = everything ok, hash didn't changed
// false = hash changed
bool CClientList::ComparePriorUserhash(uint32 dwIP, uint16 nPort, void* pNewHash){
	CDeletedClient* pResult = 0;
	if (m_trackedClientsList.Lookup(dwIP, pResult)){
		for (int i = 0; i != pResult->m_ItemsList.GetCount(); i++){
			if (pResult->m_ItemsList[i].nPort == nPort){
				// X-Ray :: ExtenedCreditTableArragement :: Start
				/*
				if (pResult->m_ItemsList[i].pHash != pNewHash)
				*/
				if (md4cmp(pResult->m_ItemsList[i].pHash, pNewHash) != 0)
				// X-Ray :: ExtenedCreditTableArragement :: End
					return false;
				else
					break;
			}
		}
	}
	return true;
}

UINT CClientList::GetClientsFromIP(uint32 dwIP) const
{
	CDeletedClient* pResult;
	if (m_trackedClientsList.Lookup(dwIP, pResult))
		return pResult->m_ItemsList.GetCount();
	return 0;
}

void CClientList::TrackBadRequest(const CUpDownClient* upcClient, int nIncreaseCounter){
	CDeletedClient* pResult = NULL;
	if (upcClient->GetIP() == 0){
		ASSERT( false );
		return;
	}
	if (m_trackedClientsList.Lookup(upcClient->GetIP(), pResult)){
		pResult->m_dwInserted = ::GetTickCount();
		pResult->m_cBadRequest += nIncreaseCounter;
	}
	else{
		CDeletedClient* ccToAdd = new CDeletedClient(upcClient);
		ccToAdd->m_cBadRequest = nIncreaseCounter;
		m_trackedClientsList.SetAt(upcClient->GetIP(), ccToAdd);
	}
}

uint32 CClientList::GetBadRequests(const CUpDownClient* upcClient) const{
	CDeletedClient* pResult = NULL;
	if (upcClient->GetIP() == 0){
		ASSERT( false );
		return 0;
	}
	if (m_trackedClientsList.Lookup(upcClient->GetIP(), pResult)){
		return pResult->m_cBadRequest;
	}
	else
		return 0;
}

void CClientList::RemoveAllTrackedClients(){
	POSITION pos = m_trackedClientsList.GetStartPosition();
	uint32 nKey;
	CDeletedClient* pResult;
	while (pos != NULL){
		m_trackedClientsList.GetNextAssoc(pos, nKey, pResult);
		m_trackedClientsList.RemoveKey(nKey);
		delete pResult;
	}
}

void CClientList::Process()
{
	///////////////////////////////////////////////////////////////////////////
	// Cleanup banned client list
	//
	const uint32 cur_tick = ::GetTickCount();
	
	// X-Ray :: Argos :: Start
	/*
	if (m_dwLastBannCleanUp + BAN_CLEANUP_TIME < cur_tick)
	*/
	if(!thePrefs.UseArgosSystem()){ // when the argos is enabled it cleans the list for its own
		if (m_dwLastBannCleanUp + thePrefs.GetArgosProcessTimeMs(true) < cur_tick){
	// X-Ray :: Argos :: End
			m_dwLastBannCleanUp = cur_tick;

			POSITION pos = m_bannedList.GetStartPosition();
			uint32 nKey;
			uint32 dwBantime;
			while (pos != NULL)
			{
				m_bannedList.GetNextAssoc( pos, nKey, dwBantime );
				// X-Ray :: Argos :: Start
				/*
				if (dwBantime + CLIENTBANTIME < cur_tick )
				*/
				if (dwBantime + thePrefs.GetArgosGenericBanTimeMs() < cur_tick)
				// X-Ray :: Argos :: End
					RemoveBannedClient(nKey);
			}
		}
	}

	///////////////////////////////////////////////////////////////////////////
	// Cleanup tracked client list
	//
	// X-Ray :: Argos :: Start
	/*
	if (m_dwLastTrackedCleanUp + TRACKED_CLEANUP_TIME < cur_tick)
	*/
	if (m_dwLastTrackedCleanUp + thePrefs.GetArgosProcessTimeMs() < cur_tick)
	// X-Ray :: Argos :: End
	{
		m_dwLastTrackedCleanUp = cur_tick;
		if (thePrefs.GetLogBannedClients())
			AddDebugLogLine(false, _T("Cleaning up TrackedClientList, %i clients on List..."), m_trackedClientsList.GetCount());
		POSITION pos = m_trackedClientsList.GetStartPosition();
		uint32 nKey;
		CDeletedClient* pResult;
		while (pos != NULL)
		{
			m_trackedClientsList.GetNextAssoc( pos, nKey, pResult );
			// X-Ray :: Argos :: Start
			/*
			if (pResult->m_dwInserted + KEEPTRACK_TIME < cur_tick ){
			*/
			if (pResult->m_dwInserted + thePrefs.GetArgosTrackTimeMs() < cur_tick ){
			// X-Ray :: Argos :: End
				m_trackedClientsList.RemoveKey(nKey);
				delete pResult;
			}
		}
		if (thePrefs.GetLogBannedClients())
			AddDebugLogLine(false, _T("...done, %i clients left on list"), m_trackedClientsList.GetCount());
	}

	///////////////////////////////////////////////////////////////////////////
	// X-Ray :: DropSystem :: Start
	//
	if((m_dwLastCleanUpDontAskThisIP + MIN2MS(1)) < cur_tick){
		m_dwLastCleanUpDontAskThisIP = cur_tick;

		POSITION pos = m_DontAskThisIPList.GetStartPosition();
		uint32 uKey;
		while (pos != NULL){
			CDontAskThisIPValues* pValue = NULL;
			m_DontAskThisIPList.GetNextAssoc( pos, uKey, pValue );
			uint8 strType = pValue->uDropType;
			uint32 uDontAskTime = pValue->uTimestamp;
			uint32 uAdditionalTime = 0;

			switch(strType){
				case 1:		uAdditionalTime = thePrefs.GetBlockTimeL2L();	break;
				case 2:		uAdditionalTime = thePrefs.GetBlockTimeUNK();	break;
				case 3:		uAdditionalTime = thePrefs.GetBlockTimeNNS();	break;
				case 4:		uAdditionalTime = thePrefs.GetBlockTimeFQS();	break;
				case 5:		uAdditionalTime = thePrefs.GetBlockTimeHQS();	break;
				case 6:		uAdditionalTime = thePrefs.GetBlockTimeTMC();	break;
				default:	break;
			}

			if((uDontAskTime + uAdditionalTime) < cur_tick){
				m_DontAskThisIPList.RemoveKey(uKey);
				delete pValue;
			}
		}
	}
	// X-Ray :: DropSystem :: End

	///////////////////////////////////////////////////////////////////////////
	// Process Kad client list
	//
	//We need to try to connect to the clients in m_KadList
	//If connected, remove them from the list and send a message back to Kad so we can send a ACK.
	//If we don't connect, we need to remove the client..
	//The sockets timeout should delete this object.
 	// X-Ray :: LoopImprovement :: Start
	/*
	POSITION pos1, pos2;
	*/
 	// X-Ray :: LoopImprovement :: End

	// buddy is just a flag that is used to make sure we are still connected or connecting to a buddy.
	buddyState buddy = Disconnected;

	// X-Ray :: LoopImprovement :: Start
	/*
	for (pos1 = m_KadList.GetHeadPosition(); (pos2 = pos1) != NULL; )
	{
		m_KadList.GetNext(pos1);
		CUpDownClient* cur_client =	m_KadList.GetAt(pos2);
	*/
	for (POSITION pos = m_KadList.GetHeadPosition(); pos != NULL;) {
		CUpDownClient* cur_client =	m_KadList.GetNext(pos);
	// X-Ray :: LoopImprovement :: End
		if( !Kademlia::CKademlia::IsRunning() )
		{
			//Clear out this list if we stop running Kad.
			//Setting the Kad state to KS_NONE causes it to be removed in the switch below.
			cur_client->SetKadState(KS_NONE);
		}
		switch(cur_client->GetKadState())
		{
			case KS_QUEUED_FWCHECK:
				//Another client asked us to try to connect to them to check their firewalled status.
				cur_client->TryToConnect(true);
				break;

			case KS_CONNECTING_FWCHECK:
				//Ignore this state as we are just waiting for results.
				break;

			case KS_CONNECTED_FWCHECK:
				//We successfully connected to the client.
				//We now send a ack to let them know.
				if (thePrefs.GetDebugClientKadUDPLevel() > 0)
					DebugSend("KADEMLIA_FIREWALLED_ACK_RES", cur_client->GetIP(), cur_client->GetKadPort());
				Kademlia::CKademlia::GetUDPListener()->SendNullPacket(KADEMLIA_FIREWALLED_ACK_RES, ntohl(cur_client->GetIP()), cur_client->GetKadPort());
				//We are done with this client. Set Kad status to KS_NONE and it will be removed in the next cycle.
				cur_client->SetKadState(KS_NONE);
				break;

			case KS_INCOMING_BUDDY:
				//A firewalled client wants us to be his buddy.
				//If we already have a buddy, we set Kad state to KS_NONE and it's removed in the next cycle.
				//If not, this client will change to KS_CONNECTED_BUDDY when it connects.
				if( m_nBuddyStatus == Connected )
					cur_client->SetKadState(KS_NONE);
				break;

			case KS_QUEUED_BUDDY:
				//We are firewalled and want to request this client to be a buddy.
				//But first we check to make sure we are not already trying another client.
				//If we are not already trying. We try to connect to this client.
				//If we are already connected to a buddy, we set this client to KS_NONE and it's removed next cycle.
				//If we are trying to connect to a buddy, we just ignore as the one we are trying may fail and we can then try this one.
				if( m_nBuddyStatus == Disconnected )
				{
					buddy = Connecting;
					m_nBuddyStatus = Connecting;
					cur_client->SetKadState(KS_CONNECTING_BUDDY);
					cur_client->TryToConnect(true);
					theApp.emuledlg->serverwnd->UpdateMyInfo();
				}
				else if( m_nBuddyStatus == Connected )
					cur_client->SetKadState(KS_NONE);
				break;

			case KS_CONNECTING_BUDDY:
				//We are trying to connect to this client.
				//Although it should NOT happen, we make sure we are not already connected to a buddy.
				//If we are we set to KS_NONE and it's removed next cycle.
				//But if we are not already connected, make sure we set the flag to connecting so we know 
				//things are working correctly.
				if( m_nBuddyStatus == Connected )
					cur_client->SetKadState(KS_NONE);
				else
				{
					ASSERT( m_nBuddyStatus == Connecting );
					buddy = Connecting;
				}
				break;

			case KS_CONNECTED_BUDDY:
				//A potential connected buddy client wanting to me in the Kad network
				//We set our flag to connected to make sure things are still working correctly.
				buddy = Connected;
				
				//If m_nBuddyStatus is not connected already, we set this client as our buddy!
				if( m_nBuddyStatus != Connected )
				{
					m_pBuddy = cur_client;
					m_nBuddyStatus = Connected;
					theApp.emuledlg->serverwnd->UpdateMyInfo();
				}
				if( m_pBuddy == cur_client && theApp.IsFirewalled() && cur_client->SendBuddyPingPong() )
				{
					if (thePrefs.GetDebugClientTCPLevel() > 0)
						DebugSend("OP__BuddyPing", cur_client);
					Packet* buddyPing = new Packet(OP_BUDDYPING, 0, OP_EMULEPROT);
					theStats.AddUpDataOverheadOther(buddyPing->size);
					cur_client->SafeSendPacket(buddyPing);
					cur_client->SetLastBuddyPingPongTime();
				}
				break;

			default:
				RemoveFromKadList(cur_client);
		}
	}
	
	//We either never had a buddy, or lost our buddy..
	if( buddy == Disconnected )
	{
		if( m_nBuddyStatus != Disconnected || m_pBuddy )
		{
			if( Kademlia::CKademlia::IsRunning() && theApp.IsFirewalled() )
			{
				//We are a lowID client and we just lost our buddy.
				//Go ahead and instantly try to find a new buddy.
				Kademlia::CKademlia::GetPrefs()->SetFindBuddy();
			}
			m_pBuddy = NULL;
			m_nBuddyStatus = Disconnected;
			theApp.emuledlg->serverwnd->UpdateMyInfo();
		}
	}

	if ( Kademlia::CKademlia::IsConnected() )
	{
		if( Kademlia::CKademlia::IsFirewalled() )
		{
			//TODO: Kad buddies won'T work with RequireCrypt, so it is disabled for now but should (and will)
			//be fixed in later version
			if( m_nBuddyStatus == Disconnected && Kademlia::CKademlia::GetPrefs()->GetFindBuddy() && !thePrefs.IsClientCryptLayerRequired())
			{
				//We are a firewalled client with no buddy. We have also waited a set time 
				//to try to avoid a false firewalled status.. So lets look for a buddy..
				if( !Kademlia::CSearchManager::PrepareLookup(Kademlia::CSearch::FINDBUDDY, true, Kademlia::CUInt128(true).Xor(Kademlia::CKademlia::GetPrefs()->GetKadID())) )
				{
					//This search ID was already going. Most likely reason is that
					//we found and lost our buddy very quickly and the last search hadn't
					//had time to be removed yet. Go ahead and set this to happen again
					//next time around.
					Kademlia::CKademlia::GetPrefs()->SetFindBuddy();
				}
			}
		}
		else
		{
			if( m_pBuddy )
			{
				//Lets make sure that if we have a buddy, they are firewalled!
				//If they are also not firewalled, then someone must have fixed their firewall or stopped saturating their line.. 
				//We just set the state of this buddy to KS_NONE and things will be cleared up with the next cycle.
				if( !m_pBuddy->HasLowID() )
					m_pBuddy->SetKadState(KS_NONE);
			}
		}
	}
	else
	{
		if( m_pBuddy )
		{
			//We are not connected anymore. Just set this buddy to KS_NONE and things will be cleared out on next cycle.
			m_pBuddy->SetKadState(KS_NONE);
		}
	}

	///////////////////////////////////////////////////////////////////////////
	// Cleanup client list
	//
	CleanUpClientList();
}

#ifdef _DEBUG
void CClientList::Debug_SocketDeleted(CClientReqSocket* deleted) const
{
	for (POSITION pos = list.GetHeadPosition(); pos != NULL;){
		CUpDownClient* cur_client =	list.GetNext(pos);
		if (!AfxIsValidAddress(cur_client, sizeof(CUpDownClient))) {
			AfxDebugBreak();
		}
		if (thePrefs.m_iDbgHeap >= 2)
			ASSERT_VALID(cur_client);
		if (cur_client->socket == deleted){
			AfxDebugBreak();
		}
	}
}
#endif

bool CClientList::IsValidClient(CUpDownClient* tocheck) const
{
	if (thePrefs.m_iDbgHeap >= 2)
		ASSERT_VALID(tocheck);
	return list.Find(tocheck)!=NULL;
}


///////////////////////////////////////////////////////////////////////////////
// Kad client list

void CClientList::RequestTCP(Kademlia::CContact* contact)
{
	uint32 nContactIP = ntohl(contact->GetIPAddress());
	// don't connect ourself
	if (theApp.serverconnect->GetLocalIP() == nContactIP && thePrefs.GetPort() == contact->GetTCPPort())
		return;

	CUpDownClient* pNewClient = FindClientByIP(nContactIP, contact->GetTCPPort());

	if (!pNewClient)
		pNewClient = new CUpDownClient(0, contact->GetTCPPort(), contact->GetIPAddress(), 0, 0, false );

	//Add client to the lists to be processed.
	pNewClient->SetKadPort(contact->GetUDPPort());
	pNewClient->SetKadState(KS_QUEUED_FWCHECK);
	m_KadList.AddTail(pNewClient);
	//This method checks if this is a dup already.
	AddClient(pNewClient);
}

void CClientList::RequestBuddy(Kademlia::CContact* contact)
{
	uint32 nContactIP = ntohl(contact->GetIPAddress());
	// don't connect ourself
	if (theApp.serverconnect->GetLocalIP() == nContactIP && thePrefs.GetPort() == contact->GetTCPPort())
		return;
	CUpDownClient* pNewClient = FindClientByIP(nContactIP, contact->GetTCPPort());
	if (!pNewClient)
		pNewClient = new CUpDownClient(0, contact->GetTCPPort(), contact->GetIPAddress(), 0, 0, false );

	//Add client to the lists to be processed.
	pNewClient->SetKadPort(contact->GetUDPPort());
	pNewClient->SetKadState(KS_QUEUED_BUDDY);
	byte ID[16];
	contact->GetClientID().ToByteArray(ID);
	pNewClient->SetUserHash(ID);
	AddToKadList(pNewClient);
	//This method checks if this is a dup already.
	AddClient(pNewClient);
}

void CClientList::IncomingBuddy(Kademlia::CContact* contact, Kademlia::CUInt128* buddyID )
{
	uint32 nContactIP = ntohl(contact->GetIPAddress());
	//If eMule already knows this client, abort this.. It could cause conflicts.
	//Although the odds of this happening is very small, it could still happen.
	if (FindClientByIP(nContactIP, contact->GetTCPPort()))
	{
		return;
	}

	// don't connect ourself
	if (theApp.serverconnect->GetLocalIP() == nContactIP && thePrefs.GetPort() == contact->GetTCPPort())
		return;

	//Add client to the lists to be processed.
	CUpDownClient* pNewClient = new CUpDownClient(0, contact->GetTCPPort(), contact->GetIPAddress(), 0, 0, false );
	pNewClient->SetKadPort(contact->GetUDPPort());
	pNewClient->SetKadState(KS_INCOMING_BUDDY);
	byte ID[16];
	contact->GetClientID().ToByteArray(ID);
	pNewClient->SetUserHash(ID);
	buddyID->ToByteArray(ID);
	pNewClient->SetBuddyID(ID);
	AddToKadList(pNewClient);
	AddClient(pNewClient);
}

void CClientList::RemoveFromKadList(CUpDownClient* torem){
	POSITION pos = m_KadList.Find(torem);
	if(pos)
	{
		if(torem == m_pBuddy)
		{
			m_pBuddy = NULL;
			theApp.emuledlg->serverwnd->UpdateMyInfo();
		}
		m_KadList.RemoveAt(pos);
	}
}

void CClientList::AddToKadList(CUpDownClient* toadd){
	if(!toadd)
		return;
	POSITION pos = m_KadList.Find(toadd);
	if(pos)
	{
		return;
	}
	m_KadList.AddTail(toadd);
}

void CClientList::CleanUpClientList(){
	// we remove clients which are not needed any more by time
	// this check is also done on CUpDownClient::Disconnected, however it will not catch all
	// cases (if a client changes the state without beeing connected
	//
	// Adding this check directly to every point where any state changes would be more effective,
	// is however not compatible with the current code, because there are points where a client has
	// no state for some code lines and the code is also not prepared that a client object gets
	// invalid while working with it (aka setting a new state)
	// so this way is just the easy and safe one to go (as long as emule is basically single threaded)
	const uint32 cur_tick = ::GetTickCount();
	if (m_dwLastClientCleanUp + CLIENTLIST_CLEANUP_TIME < cur_tick ){
		m_dwLastClientCleanUp = cur_tick;
		// X-Ray :: LoopImprovement :: Start
		/*
		POSITION pos1, pos2;
		*/
		// X-Ray :: LoopImprovement :: End
		uint32 cDeleted = 0;
		// X-Ray :: LoopImprovement :: Start
		/*
		for (pos1 = list.GetHeadPosition();( pos2 = pos1 ) != NULL;){
			list.GetNext(pos1);
			CUpDownClient* pCurClient =	list.GetAt(pos2);
		*/
		for (POSITION pos = list.GetHeadPosition(); pos;){
			CUpDownClient* pCurClient =	list.GetNext(pos);
		// X-Ray :: LoopImprovement :: End
			if ((pCurClient->GetUploadState() == US_NONE || pCurClient->GetUploadState() == US_BANNED && !pCurClient->IsBanned())
				&& pCurClient->GetDownloadState() == DS_NONE
				&& pCurClient->GetChatState() == MS_NONE
				&& pCurClient->GetKadState() == KS_NONE
				&& pCurClient->GetRequestFile() == NULL // X-Ray :: Optimizations - kill only clients without reqfile
				&& pCurClient->socket == NULL)
			{
				cDeleted++;
				delete pCurClient;
			}
		}
		DEBUG_ONLY(AddDebugLogLine(false,_T("Cleaned ClientList, removed %i not used known clients"), cDeleted));
	}
}


CDeletedClient::CDeletedClient(const CUpDownClient* pClient)
{
	m_cBadRequest = 0;
	m_dwInserted = ::GetTickCount();
	// X-Ray :: ExtenedCreditTableArragement :: Start
	/*
	PORTANDHASH porthash = { pClient->GetUserPort(), pClient->Credits()};
	*/
	// track new port & hash
	PORTANDHASH porthash;
	porthash.nPort= pClient->GetUserPort();
	md4cpy(porthash.pHash,pClient->GetUserHash());
	// X-Ray :: ExtenedCreditTableArragement :: End
	m_ItemsList.Add(porthash);
}

// ZZ:DownloadManager -->
void CClientList::ProcessA4AFClients() const {
    //if(thePrefs.GetLogA4AF()) AddDebugLogLine(false, _T(">>> Starting A4AF check"));
	// X-Ray :: LoopImprovement :: Start
	/*
	POSITION pos1, pos2;
	for (pos1 = list.GetHeadPosition();( pos2 = pos1 ) != NULL;){
		list.GetNext(pos1);
		CUpDownClient* cur_client =	list.GetAt(pos2);
	*/
	for (POSITION pos = list.GetHeadPosition(); pos;){
		CUpDownClient* cur_client =	list.GetNext(pos);
	// X-Ray :: LoopImprovement :: End

        if(cur_client->GetDownloadState() != DS_DOWNLOADING &&
           cur_client->GetDownloadState() != DS_CONNECTED &&
           (!cur_client->m_OtherRequests_list.IsEmpty() || !cur_client->m_OtherNoNeeded_list.IsEmpty())) {
            //AddDebugLogLine(false, _T("+++ ZZ:DownloadManager: Trying for better file for source: %s '%s'"), cur_client->GetUserName(), cur_client->reqfile->GetFileName());
               cur_client->SwapToAnotherFile(_T("Periodic A4AF check CClientList::ProcessA4AFClients()"), false, false, false, NULL, true, false);
        }
	}
    //if(thePrefs.GetLogA4AF()) AddDebugLogLine(false, _T(">>> Done with A4AF check"));
}
// <-- ZZ:DownloadManager

void CClientList::AddKadFirewallRequest(uint32 dwIP){
	IPANDTICS add = {dwIP, ::GetTickCount()};
	listFirewallCheckRequests.AddHead(add);
	while (!listFirewallCheckRequests.IsEmpty()){
		if (::GetTickCount() - listFirewallCheckRequests.GetTail().dwInserted > SEC2MS(180))
			listFirewallCheckRequests.RemoveTail();
		else
			break;
	}
}

bool CClientList::IsKadFirewallCheckIP(uint32 dwIP) const{
	for (POSITION pos = listFirewallCheckRequests.GetHeadPosition(); pos != NULL; listFirewallCheckRequests.GetNext(pos)){
		if (listFirewallCheckRequests.GetAt(pos).dwIP == dwIP && ::GetTickCount() - listFirewallCheckRequests.GetAt(pos).dwInserted < SEC2MS(180))
			return true;
	}
	return false;
}

// X-Ray :: DropSystem :: Start
bool CClientList::DontAskThisIP(uint32 dwIP)
{
	CDontAskThisIPValues* pValue = NULL;

	if(m_DontAskThisIPList.Lookup(dwIP, pValue)){
		uint8 uDropType = pValue->uDropType;
		uint32 uDontAskTime = pValue->uTimestamp;
		uint32 uAdditionalTime = 0;

		switch(uDropType){
			case 1:		uAdditionalTime = thePrefs.GetBlockTimeL2L();	break;
			case 2:		uAdditionalTime = thePrefs.GetBlockTimeUNK();	break;
			case 3:		uAdditionalTime = thePrefs.GetBlockTimeNNS();	break;
			case 4:		uAdditionalTime = thePrefs.GetBlockTimeFQS();	break;
			case 5:		uAdditionalTime = thePrefs.GetBlockTimeHQS();	break;
			case 6:		uAdditionalTime = thePrefs.GetBlockTimeTMC();	break;
			default:	break;
		}

		if((uDontAskTime + uAdditionalTime) > ::GetTickCount())
			return true;
	}
	return false; 
}
void CClientList::AddToDontAskThisIP(uint32 dwIP, uint8 uDropType)
{
	CDontAskThisIPValues* pValue = new CDontAskThisIPValues;
	pValue->uTimestamp = ::GetTickCount();
	pValue->uDropType = uDropType;
	m_DontAskThisIPList.SetAt(dwIP, pValue);
}
// X-Ray :: DropSystem :: End

// X-Ray :: ReAskSrcAfterIPChange :: Start
void CClientList::TrigReaskForDownload(){
	const DWORD dwCurTick = ::GetTickCount();
	const bool bLowID = theApp.IsFirewalled(); // we will not inform but do a complete reask, if we are on LowID

	for(POSITION pos = list.GetHeadPosition(); pos;){
		CUpDownClient* cur_client = list.GetNext(pos);

		const DWORD ReAskTime = cur_client->GetSpreadReAskTime() // X-Ray :: SpreadReask
								+ thePrefs.GetReAskTimeDif(); // X-Ray :: VariableReaskTime

		if(cur_client->reqfile){
			// early abort if reask soon!
			if(cur_client->GetTimeUntilReask() <= MIN2MS(2))
				continue;

			/*
			** Let's add some pseudo code here:
			** I'm setting the last asked time to a value that the client will be reasked
			** real soon
			**
			** a = time now
			** b = ReAsk time
			** c = time till MIN_REQUESTTIME
			**
			** SetLastAskedTime(a-(b-c))
			**
			** The result should be that the client will be reasked either immediatly
			** or as soon as the MIN_REQUESTTIME expired.
			**
			** Hell am I proud of this solution! *rofl*
			*/

			// eMules/eDonkeys should understand the OP_CHANGE_CLIENT_ID. So we inform them with this opcode. Other clients
			// will be reasked, as we can't always know, if they support this opcode.
			if ((cur_client->GetClientSoft() == SO_EMULE || cur_client->GetClientSoft() == SO_EMULEPLUS || cur_client->GetClientSoft() == SO_EDONKEY || cur_client->GetClientSoft() == SO_EDONKEYHYBRID) && !bLowID){ // inform
				if (thePrefs.GetDebugClientTCPLevel() > 0)
					DebugSend("OP__Change_Client_Id", cur_client);
				Packet* packet = new Packet(OP_CHANGE_CLIENT_ID, 8);
				PokeUInt32(packet->pBuffer, theApp.GetPublicIP()); // New ID
				PokeUInt32(packet->pBuffer + 4, theApp.serverconnect->IsConnected() ? theApp.serverconnect->GetCurrentServer()->GetIP() : 0x00000000); // New Server IP
				theStats.AddUpDataOverheadOther(packet->size);
				cur_client->SafeSendPacket(packet);
			} else // reask
				cur_client->SetLastAskedTime(cur_client->reqfile, dwCurTick-(ReAskTime-cur_client->GetTimeUntilReask(cur_client->reqfile, true, false, false)));
		}
	}
	return;
}
// X-Ray :: ReAskSrcAfterIPChange :: End

// X-Ray :: IP2Country :: Start
void CClientList::ResetIP2Country(){
	for(POSITION pos = list.GetHeadPosition(); pos != NULL;) {
		list.GetNext(pos)->ResetIP2Country();
	}
}
// X-Ray :: IP2Country :: End

// X-Ray :: CompatClientStats :: Start
void CClientList::GetCompatClientsStats(CRBMap<CString, uint32> *compatClients)
{
	for (POSITION pos = list.GetHeadPosition(); pos != NULL;) {		
		CUpDownClient* cur_client = list.GetNext(pos);
		uint32 count = 0;

		switch (cur_client->GetClientSoft())
		{
			case SO_HYDRANODE:
			{
				if (!compatClients->Lookup(_T("Hydranode"), count))
					count = 1;
				else
					count++;
				compatClients->SetAt(_T("Hydranode"), count);
				break;
			}
			case SO_EMULEPLUS:
			{
				if (!compatClients->Lookup(_T("eMule Plus"), count))
					count = 1;
				else
					count++;
				compatClients->SetAt(_T("eMule Plus"), count);
				break;
			}
			case SO_TRUSTYFILES:
			{
				if (!compatClients->Lookup(_T("TrustyFiles"), count))
					count = 1;
				else
					count++;
				compatClients->SetAt(_T("TrustyFiles"), count);
				break;
			}
			case SO_CDONKEY:
			{
				if (!compatClients->Lookup(_T("cDonkey"), count))
					count = 1;
				else
					count++;
				compatClients->SetAt(_T("cDonkey"), count);
				break;
			}
			case SO_XMULE:
			{
				if (!compatClients->Lookup(_T("xMule"), count))
					count = 1;
				else
					count++;
				compatClients->SetAt(_T("xMule"), count);
				break;
			}
			case SO_LPHANT:
			{
				if (!compatClients->Lookup(_T("lphant"), count))
					count = 1;
				else
					count++;
				compatClients->SetAt(_T("lphant"), count);
				break;
			}
			default:
				continue;
		}
	}
}
// X-Ray :: CompatClientStats :: End